@@ -14,9 +14,14 @@
 
 #include <ft2build.h>
 #include FT_FREETYPE_H
+#include FT_FONT_FORMATS_H
+#include FT_MODULE_H
+#include FT_TRUETYPE_DRIVER_H
+#include FT_CFF_DRIVER_H
 #include FT_MULTIPLE_MASTERS_H
 
 #include "common.h"
+#include "mlgetopt.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -32,9 +37,11 @@
 #define  MAXPTSIZE    500               /* dtp */
 #define  MAX_MM_AXES    6
 
+#define N_CFF_HINTING_ENGINES  2
 
-  static char  Header[256];
-  static char* new_header = 0;
+
+  static char   Header[256];
+  static char*  new_header = NULL;
 
   static const unsigned char*  Text = (unsigned char*)
     "The quick brown fox jumps over the lazy dog 0123456789 "
@@ -47,7 +54,14 @@
   static FT_Size       size;         /* the font size               */
   static FT_GlyphSlot  glyph;        /* the glyph slot              */
 
-  static FT_Encoding   encoding = FT_ENCODING_NONE;
+  static unsigned long  encoding = FT_ENCODING_NONE;
+
+  static unsigned int  cff_hinting_engine;
+  static unsigned int  tt_interpreter_versions[3];
+  static unsigned int  num_tt_interpreter_versions;
+  static unsigned int  tt_interpreter_version_idx;
+
+  static const char*  font_format;
 
   static FT_Error      error;        /* error returned by FreeType? */
 
@@ -75,11 +89,11 @@
 
   static int  render_mode = 1;
 
-  static FT_MM_Var  *multimaster   = NULL;
-  static FT_Fixed    design_pos   [MAX_MM_AXES];
-  static FT_Fixed    requested_pos[MAX_MM_AXES];
-  static int         requested_cnt = 0;
-  static int         used_num_axis = 0;
+  static FT_MM_Var    *multimaster   = NULL;
+  static FT_Fixed      design_pos   [MAX_MM_AXES];
+  static FT_Fixed      requested_pos[MAX_MM_AXES];
+  static unsigned int  requested_cnt = 0;
+  static unsigned int  used_num_axis = 0;
 
 
 #define DEBUGxxx
@@ -425,6 +439,8 @@
     grLn();
     grWriteln( "p, n        previous/next font" );
     grLn();
+    grWriteln( "H           cycle through hinting engines (if available)" );
+    grLn();
     grWriteln( "Up, Down    change pointsize by 1 unit" );
     grWriteln( "PgUp, PgDn  change pointsize by 10 units" );
     grLn();
@@ -448,10 +464,45 @@
   }
 
 
+  static void
+  cff_hinting_engine_change( int  delta )
+  {
+    int  new_cff_hinting_engine = 0;
+
+
+    if ( delta )
+      new_cff_hinting_engine =
+        ( (int)cff_hinting_engine +
+          delta                   +
+          N_CFF_HINTING_ENGINES   ) % N_CFF_HINTING_ENGINES;
+
+    error = FT_Property_Set( library,
+                             "cff",
+                             "hinting-engine",
+                             &new_cff_hinting_engine );
+    if ( !error )
+      cff_hinting_engine = (FT_UInt)new_cff_hinting_engine;
+  }
+
+
+  static void
+  tt_interpreter_version_change( void )
+  {
+    tt_interpreter_version_idx += 1;
+    tt_interpreter_version_idx %= num_tt_interpreter_versions;
+
+    FT_Property_Set( library,
+                     "truetype",
+                     "interpreter-version",
+                     &tt_interpreter_versions[tt_interpreter_version_idx] );
+  }
+
+
   static int
   Process_Event( grEvent*  event )
   {
-    int  i, axis;
+    int           i;
+    unsigned int  axis;
 
 
     switch ( event->key )
@@ -467,7 +518,7 @@
     /* mode keys */
 
     case grKEY( 'a' ):
-      antialias = !antialias;
+      antialias  = !antialias;
       new_header = antialias ? (char *)"anti-aliasing is now on"
                              : (char *)"anti-aliasing is now off";
       return 1;
@@ -484,7 +535,7 @@
       return (int)event->key;
 
     case grKEY( 'h' ):
-      hinted = !hinted;
+      hinted     = !hinted;
       new_header = hinted ? (char *)"glyph hinting is now active"
                           : (char *)"glyph hinting is now ignored";
       break;
@@ -495,6 +546,13 @@
                                  : (char *)"rendering test text string";
       break;
 
+    case grKEY( 'H' ):
+      if ( !strcmp( font_format, "CFF" ) )
+        cff_hinting_engine_change( 1 );
+      else if ( !strcmp( font_format, "TrueType" ) )
+        tt_interpreter_version_change();
+      break;
+
     /* MM related keys */
 
     case grKeyF1:
@@ -628,25 +686,35 @@
        * for optical size even in PS.
        */
       pos += FT_MulDiv( i, a->maximum - a->minimum, 1000 );
-      if ( pos < a->minimum ) pos = a->minimum;
-      if ( pos > a->maximum ) pos = a->maximum;
+      if ( pos < a->minimum )
+        pos = a->minimum;
+      if ( pos > a->maximum )
+        pos = a->maximum;
 
       design_pos[axis] = pos;
 
+      /* for MM fonts, round the design coordinates to integers */
+      if ( !FT_IS_SFNT( face ) )
+        design_pos[axis] &= (FT_Fixed)0xFFFF0000L;
+
       FT_Set_Var_Design_Coordinates( face, used_num_axis, design_pos );
     }
     return 1;
 
   Do_Scale:
     ptsize += i;
-    if ( ptsize < 1 )         ptsize = 1;
-    if ( ptsize > MAXPTSIZE ) ptsize = MAXPTSIZE;
+    if ( ptsize < 1 )
+      ptsize = 1;
+    if ( ptsize > MAXPTSIZE )
+      ptsize = MAXPTSIZE;
     return 1;
 
   Do_Glyph:
     Num += i;
-    if ( Num < 0 )           Num = 0;
-    if ( Num >= num_glyphs ) Num = num_glyphs - 1;
+    if ( Num < 0 )
+      Num = 0;
+    if ( Num >= num_glyphs )
+      Num = num_glyphs - 1;
     return 1;
   }
 
@@ -703,8 +771,15 @@
     int    option;
     int    file_loaded;
 
+    unsigned int  n;
+
     grEvent  event;
 
+    unsigned int  dflt_tt_interpreter_version;
+    unsigned int  versions[3] = { TT_INTERPRETER_VERSION_35,
+                                  TT_INTERPRETER_VERSION_38,
+                                  TT_INTERPRETER_VERSION_40 };
+
 
     execname = ft_basename( argv[0] );
 
@@ -713,6 +788,30 @@
     if ( error )
       PanicZ( "Could not initialize FreeType library" );
 
+    /* get the default value as compiled into FreeType */
+    FT_Property_Get( library,
+                     "cff",
+                     "hinting-engine", &cff_hinting_engine );
+
+    /* collect all available versions, then set again the default */
+    FT_Property_Get( library,
+                     "truetype",
+                     "interpreter-version", &dflt_tt_interpreter_version );
+    for ( n = 0; n < 3; n++ )
+    {
+      error = FT_Property_Set( library,
+                               "truetype",
+                               "interpreter-version", &versions[n] );
+      if ( !error )
+        tt_interpreter_versions[
+          num_tt_interpreter_versions++] = versions[n];
+      if ( versions[n] == dflt_tt_interpreter_version )
+        tt_interpreter_version_idx = n;
+    }
+    FT_Property_Set( library,
+                     "truetype",
+                     "interpreter-version", &dflt_tt_interpreter_version );
+
     while ( 1 )
     {
       option = getopt( argc, argv, "d:e:f:h:r:vw:" );
@@ -727,7 +826,7 @@
         break;
 
       case 'e':
-        encoding = (FT_Encoding)make_tag( optarg );
+        encoding = make_tag( optarg );
         break;
 
       case 'f':
@@ -792,11 +891,16 @@
     /* Load face */
     error = FT_New_Face( library, argv[file], 0, &face );
     if ( error )
+    {
+      face = NULL;
       goto Display_Font;
+    }
+
+    font_format = FT_Get_Font_Format( face );
 
     if ( encoding != FT_ENCODING_NONE )
     {
-      error = FT_Select_Charmap( face, encoding );
+      error = FT_Select_Charmap( face, (FT_Encoding)encoding );
       if ( error )
         goto Display_Font;
     }
@@ -804,32 +908,34 @@
     /* retrieve multiple master information */
     error = FT_Get_MM_Var( face, &multimaster );
     if ( error )
+    {
+      multimaster = NULL;
       goto Display_Font;
+    }
 
     /* if the user specified a position, use it, otherwise */
     /* set the current position to the median of each axis */
+    if ( multimaster->num_axis > MAX_MM_AXES )
     {
-      int  n;
+      fprintf( stderr, "only handling first %d GX axes (of %d)\n",
+                       MAX_MM_AXES, multimaster->num_axis );
+      used_num_axis = MAX_MM_AXES;
+    }
+    else
+      used_num_axis = multimaster->num_axis;
 
+    for ( n = 0; n < used_num_axis; n++ )
+    {
+      design_pos[n] = n < requested_cnt ? requested_pos[n]
+                                        : multimaster->axis[n].def;
+      if ( design_pos[n] < multimaster->axis[n].minimum )
+        design_pos[n] = multimaster->axis[n].minimum;
+      else if ( design_pos[n] > multimaster->axis[n].maximum )
+        design_pos[n] = multimaster->axis[n].maximum;
 
-      if ( multimaster->num_axis > MAX_MM_AXES )
-      {
-        fprintf( stderr, "only handling first %d GX axes (of %d)\n",
-                         MAX_MM_AXES, multimaster->num_axis );
-        used_num_axis = MAX_MM_AXES;
-      }
-      else
-        used_num_axis = (int)multimaster->num_axis;
-
-      for ( n = 0; n < used_num_axis; n++ )
-      {
-        design_pos[n] = n < requested_cnt ? requested_pos[n]
-                                          : multimaster->axis[n].def;
-        if ( design_pos[n] < multimaster->axis[n].minimum )
-          design_pos[n] = multimaster->axis[n].minimum;
-        else if ( design_pos[n] > multimaster->axis[n].maximum )
-          design_pos[n] = multimaster->axis[n].maximum;
-      }
+      /* for MM fonts, round the design coordinates to integers */
+      if ( !FT_IS_SFNT( face ) )
+        design_pos[n] &= (FT_Fixed)0xFFFF0000L;
     }
 
     error = FT_Set_Var_Design_Coordinates( face, used_num_axis, design_pos );
@@ -895,13 +1001,13 @@
           new_header = Header;
 
         grWriteCellString( &bit, 0, 0, new_header, fore_color );
-        new_header = 0;
+        new_header = NULL;
 
         sprintf( Header, "axes:" );
         {
-          int  n;
-          int  limit = used_num_axis > MAX_MM_AXES / 2 ? MAX_MM_AXES / 2
-                                                       : used_num_axis;
+          unsigned int  limit = used_num_axis > MAX_MM_AXES / 2
+                                  ? MAX_MM_AXES / 2
+                                  : used_num_axis;
 
 
           for ( n = 0; n < limit; n++ )
@@ -920,8 +1026,7 @@
 
         if ( used_num_axis > MAX_MM_AXES / 2 )
         {
-          int  n;
-          int  limit = used_num_axis;
+          unsigned int  limit = used_num_axis;
 
 
           sprintf( Header, "     " );
@@ -941,9 +1046,24 @@
           grWriteCellString( &bit, 0, 24, Header, fore_color );
         }
 
-        sprintf( Header, "at %d points, first glyph = %d",
-                         ptsize,
-                         Num );
+        {
+          unsigned int  tt_ver = tt_interpreter_versions[
+                                   tt_interpreter_version_idx];
+
+
+          sprintf( Header, "at %d points, first glyph = %d, format = %s",
+                           ptsize,
+                           Num,
+                           strcmp( font_format, "CFF" )
+                             ? ( tt_ver == TT_INTERPRETER_VERSION_35
+                                   ? "TrueType (v35)"
+                                   : ( tt_ver == TT_INTERPRETER_VERSION_38
+                                       ? "TrueType (v38)"
+                                       : "TrueType (v40)" ) )
+                             : ( cff_hinting_engine == FT_CFF_HINTING_FREETYPE
+                                   ? "CFF (FreeType)"
+                                   : "CFF (Adobe)" ) );
+        }
       }
       else
       {
@@ -980,6 +1100,15 @@
         goto NewFile;
       }
 
+      if ( key == 'H' )
+      {
+        /* enforce reloading */
+        if ( file_loaded >= 1 )
+          FT_Done_Face( face );
+
+        goto NewFile;
+      }
+
       if ( ptsize != old_ptsize )
       {
         Reset_Scale( ptsize );
